/*
 *  Copyright 2009 Eric Jiang and Daniel Kelley
 *
 *  This file is part of cdsmrm.
 *
 *  Cdsmrm is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Cdsmrm is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with cdsmrm. If not, see http://www.gnu.org/licenses/.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <vector>
#include "bitarray.h"
#include <iostream>

using namespace std;

long findPattern(bit_array_c *datav,
		bool pattern[], int pSize,	// a linear pattern and the pattern length
		int iStart = 0, int iEnd = -1);	// start and end positions for search in bits
void blankRectangle(
		int x, int y,
		int width, int height,
		bit_array_c *datav);
void output(bit_array_c *datav);
bool isImagePositive(bit_array_c *datav);
void invertImage(bit_array_c *datav);

char* filename;
char magic[3];
int h;
int w;
// variables that start with "i" are in bits
int iRowsize;
// vars that start with "y" are in bytes
long iImglen;
long yImglen;
unsigned char* data;

//
// logo patterns
//
// bool l1pattern[31] = {1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1}; // inverted
bool l1pattern[35]   = {0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,0}; 
int l1size = 35;
bool l2pattern[30] = {0,0,1,1,1,1,1,0,0,1,1,1,1,1,0,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0};
int l2size = 30;

void ShowArray(char *name, bit_array_c *ba)
{
    cout << name << ": ";
    ba->Dump(cout);
    cout << endl;
}

int main(int argc, char** argv){

	//
	// Check to make sure user has provided the input file
	//
	if(argc <= 1) {
		printf("Usage: cdsmrm infected.pbm\nInput file is overwritten.\nInput MUST have magic number P4!\n");
		return 1;
	}
	
	bool dirtyflag = false;
	filename = argv[1];
	printf("Opening %s...\n", filename);
	FILE *in = fopen(filename, "rb");
	fscanf(in, "%s\n%d\n %d\n", magic, &w, &h);
	
	//
	// Filesize calculations
	//
	iRowsize = w + (8 - (w  % 8));
	iImglen = iRowsize * h;
	yImglen = iImglen/8;

	//
	// First, read in all image data into an array
	// Then, convert array to vector
	// Use vector to initialize new bit_array_c
	//
	printf("Allocating %ld bytes of memory...\n", yImglen+1);
	data = (unsigned char*)malloc(yImglen+1);
	fread(data, yImglen, 1, in);
	fclose(in);
	std::vector<unsigned char> vn(data, data + yImglen);
	bit_array_c datav(vn, iImglen);

	//
	// Normalize the image in case it's inverted
	// The background should always be white
	//
	if(!isImagePositive(&datav)) {
		printf("Negative image detected!  Inverting.\n");
		invertImage(&datav);
		dirtyflag = true;
	}
	/*else {
		printf("Image is positive.  All is good.\n");
	}*/

	//
	// Find CDSM logo
	// Remove using blankRectangle
	//
	long result = findPattern(&datav, l1pattern, l1size);
	if(result != -1) {
		int x = result & 0xFFFF; int y = result >> sizeof(int)*4;
		printf("Large CDSM logo found at %d, %d\n", x, y);
		blankRectangle(x - 205, y - 88, 369, 108, &datav);
		dirtyflag = true;
	}
	else {
		result = findPattern(&datav, l2pattern, l2size, iImglen >> 1); // search bottom half
		if(result != -1) {
			int x = result & 0xFFFF; int y = result >> sizeof(int)*4;
			printf("Small CDSM logo found at %d, %d\n", x, y);
			blankRectangle(x - 123, y - 53, 221, 65, &datav);
			dirtyflag = true;
		}
	}

	//
	// Write output
	//
	if(dirtyflag == true) {
		printf("Writing output...\n");
		output(&datav);
	}

	return 0;
}	

/**
 * findPattern
 *
 * @param bit_array_c*	bit array of bi-level image
 * @param pattern[]   	bit array of linear pattern
 * @param pSize       	length of pattern
 * @param iStart      	starting location for search
 * @param iEnd        	ending location for search
 *
 * @return long int with x in upper 2 bytes, y in lower 2 bytes
 */
long findPattern(bit_array_c *datav, bool pattern[], int pSize, int iStart, int iEnd){
	long row = -1;
	int column = -1;
	//
	// if iEnd is invalid or before iStart,
	// set it to the very end of the bit array
	//
	iEnd = ( iEnd < iStart	? iImglen - pSize	: iEnd );

	//
	// At each step of the outer loop, match pattern against current pixel
	// At each step of the inner loop, match each bit of pattern
	//
	for (long i = iStart; i < iEnd; i++){
		for (int t = 0; t < pSize; t++){
			// don't wraparound
			if((i+t) % iRowsize == 0) {
				break;
			}
			// bit array compare
			if ((*datav)[i + t] == pattern[t]){
				//
				// If we've matched to the end of the pattern
				// then we've succeeded
				//
				if(t == pSize - 1) {
					row = i / iRowsize;
					column = i % iRowsize;
					// 0xROWWclmn
					return (row << (sizeof(int)*4) )| column ;
				}
			}
			// pattern fails to hold
			else {
				break;
			}
		}
	}
	return -1;
}

void blankRectangle(int x, int y, int width, int height, bit_array_c *datav){
	//printf("Blanking from %d,%d to %d,%d\n", x, y, x+width, y+height);
	long bit = 0; // ints overflow??
	for (int yy = y; yy < y + height; yy++){
		for (int xx = x; xx < x + width; xx++){
			(*datav).ClearBit(yy*iRowsize + xx);
		}
	}
}

void output(bit_array_c *datav){
	FILE* out = fopen(filename, "wb");
	if (out != NULL){
		fprintf(out, "%s\n%d %d\n", "P4", w, h);
		vector<unsigned char>::iterator it;
		for (it = (*datav).m_Array.begin(); it < (*datav).m_Array.end(); it++){
			fputc(*it, out);
		}
		fclose(out);
	}
}

/**
 * isImagePositive
 *
 * Tests to see if the image is positive.  That is, if the image
 * has a white background.  According to the Netpbm documentation,
 * boolean 1 is black and 0 is white.  However, pdfimages thinks
 * otherwise.
 *
 * So, this function tests a bunch of pixels near the beginning
 * (top) of the image and sees that if more are white, then it's
 * positive, and if more are black, than it's negative.
 *
 */
bool isImagePositive(bit_array_c *datav) {
#define iSkip	1000
#define tests	10
	int numWhite = 0;
	int numBlack = 0;
	for (int i = 0; i < tests; i++) {
		if((*datav)[(i+1)*iSkip] == 1) {
			numBlack++;
		}
		else {
			numWhite++;
		}
	}
	return (numWhite > numBlack);
}

/**
 * invertImage
 *
 * inverts the given bit array one char at a time.
 *
 */
void invertImage(bit_array_c *datav) {
	for(int i = 0; i < (*datav).m_Array.size(); i++) {
		(*datav).m_Array[i] = ~ (*datav).m_Array[i];
	}
}
